/*------------------------------------------------------------------------------------
UserCompMethods.c

This file contains the definition of several comparison methods

Copyright: Universit de Savoie, LISTIC/ESIA.

Author: L. Foulloy

Revision: V3.0 - L. Foulloy - 08/06/2003
------------------------------------------------------------------------------------ */

#include "simstruc.h"
#include "type.h"

//---------------------------------------------------------------------------------------------------------
// Constants for Debuging
//
//#define DEBUG_3

//----------------------------------------------------------------------------------------------------------
// Prototypes
//
static Trapeze DiffObjectiveMeasure(SimStruct * S);
static double Sign(SimStruct * S);
static double Pos(SimStruct * S);
static double Nec(SimStruct * S);
static double Haussdorf(Interval A, Interval B);
static Interval AlphaCut(Trapeze t, double alpha);
static Trapeze CompHaussdorf(SimStruct * S);

//----------------------------------------------------------------------------------------------------------
//   Comparison method                                               
//
void Comp(SimStruct * S)
{
    int_T *					IWork = ssGetIWork(S);
    real_T * 				u = ssGetOutputPortRealSignal(S,0);

  switch(IWork[0])
    {
    case 0 : // Comparison using the fuzzy difference between the support of the objective and the support of the measurement.
      *((Trapeze *)u)= DiffObjectiveMeasure(S);
      return;
    case 1 : // Comparison using the "signed possibility". The result is a scalar transformed into a singleton. 
      u[0] = Sign(S)*(1-Pos(S));
      u[1] = u[0];
      u[2] = u[0];
      u[3] = u[0];
      return;
    case 2 : // Comparison using the "signed necessity". The result is a scalar transformed into a singleton. 
      u[0] = Sign(S)*(1-Nec(S));
      u[1] = u[0];
      u[2] = u[0];
      u[3] = u[0];
      return;
    case 3 : // Comparison using Haussdorf's distance.
      *((Trapeze *)u)= CompHaussdorf(S);
      return;
    case 4 : // Unsigned possibility. The result is a scalar transformed into a singleton.
      u[0] = Pos(S);
      u[1] = u[0];
      u[2] = u[0];
      u[3] = u[0];
      return;
    case 5 : // Unsigned necessity. The result is a scalar transformed into a singleton.
      u[0] = Nec(S);
      u[1] = u[0];
      u[2] = u[0];
      u[3] = u[0];
      return;

   default: ERROR(ERR_UNKNOWN_COMP);return;
    }
}

//----------------------------------------------------------------------------------------------------------
//                   Difference between the objective and the measurement
//
static Trapeze DiffObjectiveMeasure(SimStruct * S)
{
    InputRealPtrsType 	Objective = ssGetInputPortRealSignalPtrs(S,0);
    InputRealPtrsType 	Measure   = ssGetInputPortRealSignalPtrs(S,1);
    double 		g1,g2,mg1,mg2,md1,md2,d1,d2;
    Trapeze		Result;

  g1  = *Measure[0];
  mg1 = *Measure[1];
  md1 = *Measure[2];
  d1  = *Measure[3];

  g2  = *Objective[0];
  mg2 = *Objective[1];
  md2 = *Objective[2];
  d2  = *Objective[3];

  Result.Left = g2-d1;
  Result.LeftMode = mg2-md1;
  Result.RightMode = md2-mg1;
  Result.Right = d2-g1;
  return(Result); 
}

//----------------------------------------------------------------------------------------------------------
//                   Sign computed from the kernels                        
//
static double Sign(SimStruct * S)
{
    InputRealPtrsType 	Objective = ssGetInputPortRealSignalPtrs(S,0);
    InputRealPtrsType 	Measure   = ssGetInputPortRealSignalPtrs(S,1);
    double 					mg1,mg2,md1,md2;

  mg1 = *Measure[1];
  md1 = *Measure[2];

  mg2 = *Objective[1];
  md2 = *Objective[2];

  if(mg1+md1 <= mg2+md2)
    return(1.0);
  else
    return(-1.0);
}

//----------------------------------------------------------------------------------------------------------
//                   Possibility degree                          
//
static double Pos(SimStruct * S)

{
    InputRealPtrsType 	Objective = ssGetInputPortRealSignalPtrs(S,0);
    InputRealPtrsType 	Measure   = ssGetInputPortRealSignalPtrs(S,1);

    double 					x;
    double 					g1,g2,mg1,mg2,md1,md2,d1,d2;
    GradeOfMembership 	Result;

  g1  = *Measure[0];
  mg1 = *Measure[1];
  md1 = *Measure[2];
  d1  = *Measure[3];

  g2  = *Objective[0];

  mg2 = *Objective[1];
  md2 = *Objective[2];
  d2  = *Objective[3];

#ifdef DEBUG_3
printf("\n");
printf("M %f %f %f %f\n",g1,mg1,md1,d1);
printf("O %f %f %f %f\n",g2,mg2,md2,d2);
#endif

  Result = 0;

  // + = Inscreasing part of the membership function
  // - = Decreasing part of the membership function

  // Case where the kernels are intersecting
  if((mg2 >= mg1 && mg2 <= md1) || (md2 >= mg1 && md2 <= md1))
    return(1.0);

  // Intersection Input+ Meaning+
  if(mg2==g2)
    {
    if(g1 <= mg2 && mg2 <=mg1)
      Result = Max(Result,(g2-g1)/(mg1-g1));
    }
  else 
    {
    if(mg1==g1)
      x = g1;
    else 
      x = ((mg2 - g2)*g1 - (mg1 - g1)*g2)/((mg2 -g2) - (mg1-g1));
    if((mg2 -g2) - (mg1-g1) == 0) // co-linearity
      {
      if(g1==g2 && mg1==mg2) // segments are identical
        return(1);
      }
    else
      if(g2 <= x && x <= mg2)
        Result = Max(Result,(x-g2)/(mg2-g2));
    }

#ifdef DEBUG_3
printf("Result after I+ M+ = %f\n",Result);
#endif

  // Intersection Input+ Meaning-
  if(md2==d2)
    {
    if(g1 <= md2 && md2 <=mg1)
      Result = Max(Result,(d2-g1)/(mg1-g1));
    }
  else 
    {
    if(mg1==g1)
      x = g1;
    else      
      x = ((md2 - d2)*g1 - (mg1 - g1)*d2)/((md2 -d2) - (mg1-g1));
    if(md2 <= x && x <= d2)
      Result = Max(Result,(x-d2)/(md2-d2));
    }

#ifdef DEBUG_3
printf("Result after I+ M- = %f\n",Result);
#endif

  // Intersection Input- Meaning+
  if(mg2==g2)
    {
    if(md1 <= mg2 && mg2 <=d1)
      Result = Max(Result,(g2-d1)/(md1-d1));
    }
  else 
    {
    if(md1==d1)
      x = d1;
    else 
      x = ((mg2 - g2)*d1 - (md1 - d1)*g2)/((mg2 -g2) - (md1-d1));
    if(g2 <= x && x <= mg2)
      Result = Max(Result,(x-g2)/(mg2-g2));
    }

#ifdef DEBUG_3
printf("Result after I- M+ = %f\n",Result);
#endif

  // Intersection Input- Meaning-
  if(md2==d2)
    {
    if(md1 <= md2 && md2 <=d1)
      Result = Max(Result,(d2-d1)/(md1-d1));
    }
  else 
    {
    if(md1==d1)
      x = d1;
    else 
      x = ((md2 - d2)*d1 - (md1 - d1)*d2)/((md2 -d2) - (md1-d1));
    if((md2 -d2) - (md1-d1) == 0) // co-linearity
      {
      if(d1==d2 && md1==md2) // segments are identical
        return(1);
      }
    else
      if(md2 <= x && x <= d2)
        Result = Max(Result,(x-d2)/(md2-d2));
    }

#ifdef DEBUG_3
printf("Result after I- M- = %f\n",Result);
#endif

  return(Result);
}

//----------------------------------------------------------------------------------------------------------
//                   Necessity degree                         
//
static double Nec(SimStruct * S)
{
    InputRealPtrsType 	Objective = ssGetInputPortRealSignalPtrs(S,0);
    InputRealPtrsType 	Measure   = ssGetInputPortRealSignalPtrs(S,1);
    double 					x;
    double 					g1,g2,mg1,mg2,md1,md2,d1,d2;
    GradeOfMembership 	Result;


  g1  = *Measure[0];
  mg1 = *Measure[1];
  md1 = *Measure[2];
  d1  = *Measure[3];

  g2  = *Objective[0];
  mg2 = *Objective[1];
  md2 = *Objective[2];
  d2  = *Objective[3];


#ifdef DEBUG_3
printf("\n");
printf("M %f %f %f %f\n",g1,mg1,md1,d1);
printf("O %f %f %f %f\n",g2,mg2,md2,d2);
#endif

  Result = 1;

  // + = Inscreasing part of the membership function
  // - = Decreasing part of the membership function

  // Kernel of the input has no intersection with the support of the fuzzy meaning : necessity = 0
  if(mg1>d2 || md1<g2)
    return(0);

  // Support of input is included in the kernel of the fuzzy meaning : necessity = 1
  if(mg2<=g1 && d1<=md2)
    return(1.0);

  // Intersection Input+ Meaning+
  if(mg2==g2)
    {
    if(g1 <= mg2 && mg2 <=mg1)
      Result = Min(Result,(g2-mg1)/(mg1-g1));
    }
  else 
    {
    if(mg1==g1)
      x = g1;
    else 
      x = ((mg2 - g2)*mg1 + (mg1 - g1)*g2)/((mg2 -g2) + (mg1-g1));
    if(g2 <= x && x <= mg2)
      Result = Min(Result,(x-g2)/(mg2-g2));
    }

#ifdef DEBUG_3
printf("Result after I+ M+ = %f\n",Result);
#endif

  // Intersection Input+ Meaning-
  if(md2==d2)
    {
    if(g1 <= md2 && md2 <=mg1)
      Result = Min(Result,(d2-mg1)/(mg1-g1));
    }
  else 
    {
    if(mg1==g1)
      x = g1;
    else      
      x = ((md2 - d2)*mg1 + (mg1 - g1)*d2)/((md2 -d2) + (mg1-g1));
    if((md2 -d2) + (mg1-g1) == 0) // co-linearity
      {
      if(g1==md2 && mg1==d2) // segments are identical
        return(0);
      }
    else
      if(md2 <= x && x <= d2)
        Result = Min(Result,(x-d2)/(md2-d2));
    }

#ifdef DEBUG_3
printf("Result after I+ M- = %f\n",Result);
#endif

  // Intersection Input- Meaning+
  if(mg2==g2)
    {
    if(md1 <= mg2 && mg2 <=d1)
      Result = Min(Result,(g2-md1)/(d1-md1));
    }
  else 
    {
    if(md1==d1)
      x = d1;
    else 
      x = ((mg2 - g2)*md1 + (md1 - d1)*g2)/((mg2 -g2) + (md1-d1));
    if((mg2 -g2) + (md1-d1) == 0) // co-linearity
      {
      if(g2==md1 && mg2==d1) // segments are identical
        return(0);
      }
    else
     if(g2 <= x && x <= mg2)
       Result = Min(Result,(x-g2)/(mg2-g2));
    }

#ifdef DEBUG_3
printf("Result after I- M+ = %f\n",Result);
#endif

 // Intersection Input- Meaning-
  if(md2==d2)
    {
    if(md1 <= md2 && md2 <=d1)
      Result = Min(Result,(d2-md1)/(d1-md1));
    }
  else 
    {
    if(md1==d1)
      x = d1;
    else 
      x = ((md2 - d2)*md1 + (md1 - d1)*d2)/((md2 -d2) + (md1-d1));
    if(md2 <= x && x <= d2)
      Result = Min(Result,(x-d2)/(md2-d2));
    }

#ifdef DEBUG_3
printf("Result after I- M- = %f\n",Result);
#endif

  return(Result);
}


//----------------------------------------------------------------------------------------------------------
//   Haussdorf's distance between two intervals                                            
//
static double Haussdorf(Interval A, Interval B)
{
double x;

x = Max(fabs(A.Left - B.Left),fabs(A.Right - B.Right));
return(x);
}
 
//----------------------------------------------------------------------------------------------------------
//   Alpha-cut of a fuzzy subset                                             
//
static Interval AlphaCut(Trapeze t, double alpha)
{
   Interval Result;

  Result.Left = alpha*(t.LeftMode - t.Left) + t.Left;
  Result.Right = alpha*(t.RightMode - t.Right) + t.Right;

  return(Result);
}

//----------------------------------------------------------------------------------------------------------
//       Comparison using Haussdorf's distance on the alpha-cuts
//                                            
static Trapeze CompHaussdorf(SimStruct * S)
{
    Trapeze Result;
    InputRealPtrsType 	A = ssGetInputPortRealSignalPtrs(S,0);
    InputRealPtrsType 	B = ssGetInputPortRealSignalPtrs(S,1);

  Result.LeftMode = Sign(S)*Haussdorf(AlphaCut(*((Trapeze *)*A),1),AlphaCut(*((Trapeze *)*B),1));
  Result.RightMode = Result.LeftMode;
  if(Sign(S) >0)
    {
    Result.Left = Result.LeftMode;
    Result.Right = Sign(S)*Haussdorf(AlphaCut(*((Trapeze *)*A),0),AlphaCut(*((Trapeze *)*B),0));
    }
  else  
    {
    Result.Right = Result.RightMode;
    Result.Left = Sign(S)*Haussdorf(AlphaCut(*((Trapeze *)*A),0),AlphaCut(*((Trapeze *)*B),0));
    }
  return(Result);
}

