/*------------------------------------------------------------------------------------
UserDefuzMethods.c

This file contains the definition of several defuzzification methods

Copyright: Universit de Savoie, LISTIC/ESIA.

Author: L. Foulloy

Revision: V3.0 - L. Foulloy - 08/06/2003
------------------------------------------------------------------------------------ */

#include "simstruc.h"
#include "type.h"

//----------------------------------------------------------------------------------------------------------
// Prototypes
//
static Interval AlphaCut(Trapeze t, double alpha);
static real_T DefuzHeight(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x);
static real_T DefuzSumOfCenter(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x);
static real_T DefuzSmallestOfMaximum(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x);
static real_T DefuzLargestOfMaximum(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x);
static real_T DefuzMeanOfMaximum(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x);
static real_T DefuzCenterOfGravity(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x);
static GradeOfMembership Membership (double x, FuzzyMeaning * s);

#define PROJECT Project->Defuzzification

//----------------------------------------------------------------------------------------------------------
//   Calls the defuzzification method                                                 
//
real_T Defuz(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x)
{

  switch(PROJECT.TypeDefuz)
    {
    case 0 : return(DefuzHeight(S,Project,x)); 
    case 1 : return(DefuzSumOfCenter(S,Project,x)); 
    case 2 : return(DefuzSmallestOfMaximum(S,Project,x)); 
    case 3 : return(DefuzLargestOfMaximum(S,Project,x)); 
    case 4 : return(DefuzMeanOfMaximum(S,Project,x)); 
    case 5 : return(DefuzCenterOfGravity(S,Project,x)); 

    default: ERROR(ERR_UNKNOWN_DEFUZ);return(0);
    }
}

//----------------------------------------------------------------------------------------------------------
// Height method
//
static real_T DefuzHeight(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x)
{
    int 							i;
    real_T 						Num;
    real_T						Den;

  Num = 0;
  Den = 0;
  for(i=0;i<PROJECT.FuzzyPartition.NbSymbol; i++)
     {
     Num += (*x[i])*PROJECT.TabAttribute[i].Modal; 
     Den += (*x[i]);
     }

   if (Den == 0)
      {
      ERROR(ERR_NO_RULE_ACTIVE);
      return(0);
      }

  return (Num / Den);
}

//----------------------------------------------------------------------------------------------------------
// Sum of Centers method
//
// This method uses the center of gravity and the area of the membership functions weighted by the grade of 
// membership that is :
// 
// u = Sum_i (alpha(i)*cog(i)* area(i))/Sum_i (alpha(i)* area(i))
//
// Warning : this method IS NOT the center of gravity method even if the result is close to it.
//
static real_T DefuzSumOfCenter(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x)
{
    int 							i;
    real_T 						Num;
    real_T						Den;

  Num = 0;
  Den = 0;
  for(i=0; i<PROJECT.FuzzyPartition.NbSymbol; i++)
     {
     Num += (*x[i])* PROJECT.TabAttribute[i].Cog * PROJECT.TabAttribute[i].Area; 
     Den += (*x[i])* PROJECT.TabAttribute[i].Area;
     }

   if (Den == 0)
      {
      ERROR(ERR_NO_RULE_ACTIVE);
      return(0);
      }
  return (Num / Den);
}

//----------------------------------------------------------------------------------------------------------
//   Computes the alpha-cut of a fuzzy subset                                             
//
static Interval AlphaCut(Trapeze t, double alpha)
{
   Interval Result;

  Result.Left = alpha*(t.LeftMode - t.Left) + t.Left;
  Result.Right = alpha*(t.RightMode - t.Right) + t.Right;

  return(Result);
}

//----------------------------------------------------------------------------------------------------------
// Smallest Of Maximum method
//
static real_T DefuzSmallestOfMaximum(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x)
{
    int 							i;
    real_T						Res;
    int  						Maximum;


  // Search for the maximum
  Maximum = 0;
  for(i=1; i<PROJECT.FuzzyPartition.NbSymbol; i++)
     if((*x[i]) > (*x[Maximum])) 
       Maximum = i;
  
  // Compute the lower bound of the alpha-cut of the maximum
  Res = AlphaCut(PROJECT.FuzzyPartition.TabFuzzyMeaning[Maximum].MembershipFunction.Trapeze, (*x[Maximum])).Left;

  // Search for other maximums
  for(i=0; i<PROJECT.FuzzyPartition.NbSymbol; i++)
     if(i != Maximum && (*x[i]) == (*x[Maximum])) 
       Res = Min(Res,AlphaCut(PROJECT.FuzzyPartition.TabFuzzyMeaning[i].MembershipFunction.Trapeze, (*x[i])).Left);

  return (Res);
}

//----------------------------------------------------------------------------------------------------------
// Largest Of Maximum method
//
static real_T DefuzLargestOfMaximum(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x)
{
    int 							i;
    real_T						Res;
    int  						Maximum;


  // Search for the maximum
  Maximum = 0;
  for(i=1; i<PROJECT.FuzzyPartition.NbSymbol; i++)
     if((*x[i]) > (*x[Maximum])) 
       Maximum = i;
  
  // Compute the upper bound of the alpha-cut of the maximum
  Res = AlphaCut(PROJECT.FuzzyPartition.TabFuzzyMeaning[Maximum].MembershipFunction.Trapeze, (*x[Maximum])).Right;

  // Search for other maximums
  for(i=0; i<PROJECT.FuzzyPartition.NbSymbol; i++)
     if(i != Maximum && (*x[i]) == (*x[Maximum])) 
       Res = Max(Res,AlphaCut(PROJECT.FuzzyPartition.TabFuzzyMeaning[i].MembershipFunction.Trapeze, (*x[i])).Right);

  return (Res);
}


//----------------------------------------------------------------------------------------------------------
// Mean Of Maximum method
//
static real_T DefuzMeanOfMaximum(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x)
{
    int 							i;
    real_T						Res;
    int  						Maximum;


  // Search for the maximum
  Maximum = 0;
  for(i=1; i<PROJECT.FuzzyPartition.NbSymbol; i++)
     if((*x[i]) > (*x[Maximum])) 
       Maximum = i;
  
  // Compute the mean of the alpha-cut of the maximum
  Res = (AlphaCut(PROJECT.FuzzyPartition.TabFuzzyMeaning[Maximum].MembershipFunction.Trapeze, (*x[Maximum])).Right +
         AlphaCut(PROJECT.FuzzyPartition.TabFuzzyMeaning[Maximum].MembershipFunction.Trapeze, (*x[Maximum])).Left)/2;

  // Search for other maximums
  for(i=0; i<PROJECT.FuzzyPartition.NbSymbol; i++)
     if(i != Maximum && (*x[i]) == (*x[Maximum])) 
       Res = Res/2 + (AlphaCut(PROJECT.FuzzyPartition.TabFuzzyMeaning[i].MembershipFunction.Trapeze, (*x[i])).Right +
                      AlphaCut(PROJECT.FuzzyPartition.TabFuzzyMeaning[i].MembershipFunction.Trapeze, (*x[i])).Left)/4;

  return (Res);
}

//----------------------------------------------------------------------------------------------------------
// Center of Gravity method
//
static real_T DefuzCenterOfGravity(SimStruct * S, FuzzyProject * Project, InputRealPtrsType x)
{
    int 							i;
    double						u;
    double 						Num;
    double                 Den;
    double 						Mu;

  Num = 0;
  Den = 0;

  for(u = PROJECT.GeneralAttribute.Min; u <= PROJECT.GeneralAttribute.Max; u+= PROJECT.GeneralAttribute.Step)
     {
     // Numerical reconstruction of the output membership function
     Mu = 0;
     for (i=0; i<PROJECT.FuzzyPartition.NbSymbol; i++)
        Mu = Max(Mu, Min(*x[i], Membership (u, &PROJECT.FuzzyPartition.TabFuzzyMeaning[i])));

     Num = Num + u * Mu;
     Den = Den + Mu;
     }

  return (Num/Den);    
}

//----------------------------------------------------------------------------------------------------------
// GradeOfMembership Membership (double x, FuzzyMeaning * s)
//
// Computes the grade of membership
//

static GradeOfMembership Membership (double x, FuzzyMeaning * s)
{
  if(x < s->MembershipFunction.Trapeze.Left)
    return (0.0);

  if (x <= s->MembershipFunction.Trapeze.LeftMode)
    if (s->MembershipFunction.Trapeze.LeftMode != s->MembershipFunction.Trapeze.Left)
      return ( ( x - s->MembershipFunction.Trapeze.Left ) /
	       ( s->MembershipFunction.Trapeze.LeftMode - s->MembershipFunction.Trapeze.Left ) );
    else
      return (1.0);

 if (x <= s->MembershipFunction.Trapeze.RightMode)
    return (1.0);

  if (x <= s->MembershipFunction.Trapeze.Right)
    if ( s->MembershipFunction.Trapeze.Right != s->MembershipFunction.Trapeze.RightMode )
      return ( ( s->MembershipFunction.Trapeze.Right - x ) /
	       ( s->MembershipFunction.Trapeze.Right - s->MembershipFunction.Trapeze.RightMode ) );
    else
      return (1.0);

  return (0.0);
}





